/*****************************************************/
/* ShellyV1.6: The ShellShapeGenerator by:           */
/*         RANDi                                     */
/*              (rschultz@informatik.uni-rostock.de) */
/* 21.11.1994                                        */
/*****************************************************/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>

#include "shelly.h"


struct glob
  {
    int countw, counth, countp, countl, countt;
    int calledwrtx3dl, count;
    int calledwrtt3dl;
    int calledwrtrpll;
    double N, N2, N3, L, L2, L3, W1, W2, W12, W22, W13, W23, P, P2, P3,
      Off2, Off3;
    double biggest;
    double my, a, b, A, beta, phi, omega, alpha, Scale;
    double Scans, Scano;
    double smin, smax, sd;
    double omin, omax, od;
    struct patch *patchanchor;
    double rnod;
  }
gl;

double camx = 0, camy = 500, camz = 500;
int camposspecified = NO;
struct lmerk *bigllist;
struct tmerk *bigtlist;

/********************************************/
/*cot:                                      */
/********************************************/
double
cot (double in)
{
  if (sin (in) == 0)
    return (0);
  else
    return (cos (in) / sin (in));
}


/********************************************/
/*round:                                    */
/********************************************/
double
round (double in)
{
  if (in - floor (in) < 0.5)
    return (floor (in));
  else
    return (floor (in) + 1);
}


/********************************************/
/*myfree:                                   */
/********************************************/
int
myfree (struct punkt *anker1)
{
  struct punkt *hp, *hpnew;
  if (anker1 == NULL)
    return (0);

  hp = anker1;
  hpnew = hp;
  do
    {
      hp = hpnew;
      hpnew = (*hp).next;

      free (hp);
    }
  while (hpnew != NULL);

  return (0);
}


/********************************************/
/*mycopystr:                                */
/********************************************/
char *
mycopystr (char *source, char *destination, char endchar)
{
  int ende = 1;
  char *merk;

  merk = destination;		/* merk points now on destination */
  if (source != NULL)
    while ((*source != '\0') && (*source != endchar))
      {
	*destination = *source;
	destination++;
	source++;
      }				/* while */
  *destination = '\0';
  return (merk);
}


/********************************************/
/*filecopy:                                 */
/********************************************/
void
filecopy (FILE * source, FILE * dest)
{
  int a, b;

  fseek (source, 0, 0);
  do
    {
      a = fgetc (source);
      if (a != EOF)
	{
	  b = fputc (a, dest);
	  if (b == EOF)
	    {
	      fclose (source);
	      fclose (dest);
	      exit (5);
	    };			/* if */
	};			/* if */
    }
  while (a != EOF);
}


/********************************************/
/*crosspro:                                 */
/********************************************/
void
crosspro (double a[3],double b[3],double *(c[3]))
{
 *(c[0]) = (a[1]*b[2] - a[2]*b[1]); 
 *(c[1]) = (a[2]*b[0] - a[0]*b[2]); 
 *(c[2]) = (a[0]*b[1] - a[1]*b[0]); 
} /* crosspro */ 

/********************************************/
/*ReadInfile:                               */
/********************************************/
void
ReadInfile (struct ShellyArguments *ShellyArgs, char *fin)
{
  char readline[laenge], hilf[laenge], *hp;
  FILE *fp;

  /* set some defaults */

  (*ShellyArgs).output = POV;
  (*ShellyArgs).mode = NORMAL;
  (*ShellyArgs).Scano = 0.05;
  (*ShellyArgs).Scans = 0.05;
  (*ShellyArgs).heightdiffo = 0.1;
  (*ShellyArgs).heightdiffs = 0.1;

  (*ShellyArgs).Scale = 1.0;
  (*ShellyArgs).Render = NO;
  (*ShellyArgs).N2 = 0;
  (*ShellyArgs).N3 = 0;
  (*ShellyArgs).N = 0;
  (*ShellyArgs).anim = NO;
  (*ShellyArgs).steps = 10;

  mycopystr ("-f +d +w200 +h160", (*ShellyArgs).povargs, '\0');


  fp = fopen (fin, "r");
  if (fp != NULL)
    {
      fprintf (stdout, "Parsing:");
      while (fgets (readline, laenge, fp) != NULL)
	{
	  fprintf (stdout, ".");

	  hp = strstr (readline, "alpha:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 6, hilf, '\0');
	      (*ShellyArgs).alpha = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "beta:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).beta = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "phi:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 4, hilf, '\0');
	      (*ShellyArgs).phi = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "my:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).my = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "omega:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 6, hilf, '\0');
	      (*ShellyArgs).omega = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "smin:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).smin = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "smax:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).smax = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "sd:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).sd = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "A:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 2, hilf, '\0');
	      (*ShellyArgs).A = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "a:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 2, hilf, '\0');
	      (*ShellyArgs).a = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "b:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 2, hilf, '\0');
	      (*ShellyArgs).b = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "P:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 2, hilf, '\0');
	      (*ShellyArgs).P = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "W1:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).W1 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "W2:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).W2 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "N:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 2, hilf, '\0');
	      (*ShellyArgs).N = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "L:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 2, hilf, '\0');
	      (*ShellyArgs).L = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "omin:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).omin = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "omax:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).omax = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "od:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).od = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "Hdo:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 4, hilf, '\0');
	      (*ShellyArgs).heightdiffo = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "Hds:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 4, hilf, '\0');
	      (*ShellyArgs).heightdiffs = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "Scans:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 6, hilf, '\0');
	      (*ShellyArgs).Scans = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "Scano:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 6, hilf, '\0');
	      (*ShellyArgs).Scano = atof (hilf);
	    }			/* if */

	  hp = strstr (readline, "Scale:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 6, hilf, '\0');
	      (*ShellyArgs).Scale = atof (hilf);
	    }			/* if */

	  hp = strstr (readline, "POV");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).output = POV;
	    }			/* if */
	  hp = strstr (readline, "BEZ");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).output = BEZ;
	    }			/* if */
	  hp = strstr (readline, "RPL");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).output = RPL;
	    }			/* if */
	  hp = strstr (readline, "T3D");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).output = T3D;
	    }			/* if */
	  hp = strstr (readline, "RAW");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).output = RAW;
	    }			/* if */
	  hp = strstr (readline, "X3D");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).output = X3D;
	    }			/* if */

	  hp = strstr (readline, "NORMAL");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).mode = NORMAL;
	    }			/* if */
	  hp = strstr (readline, "NODULE");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).mode = NODULE;
	    }			/* if */
	  hp = strstr (readline, "NEWNOD");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).mode = NEWNOD;
	    }			/* if */
	  hp = strstr (readline, "RENDER");
	  if (hp != NULL)
	    {
	      (*ShellyArgs).Render = YES;
	    }			/* if */

	  hp = strstr (readline, "POVARGS:");
	  if (hp != NULL)
	    mycopystr (hp + 8, (*ShellyArgs).povargs, '\n');

	  hp = strstr (readline, "P2:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).P2 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "W12:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 4, hilf, '\0');
	      (*ShellyArgs).W12 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "W22:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 4, hilf, '\0');
	      (*ShellyArgs).W22 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "N2:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).N2 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "L2:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).L2 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "P3:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).P3 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "W13:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 4, hilf, '\0');
	      (*ShellyArgs).W13 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "W23:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 4, hilf, '\0');
	      (*ShellyArgs).W23 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "N3:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).N3 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "L3:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 3, hilf, '\0');
	      (*ShellyArgs).L3 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "Off2:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).Off2 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "Off3:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).Off3 = atof (hilf);
	    }			/* if */
	  hp = strstr (readline, "camx:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      camx = atof (hilf);
	      camposspecified = YES;
	    }			/* if */
	  hp = strstr (readline, "camy:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      camy = atof (hilf);
	      camposspecified = YES;
	    }			/* if */
	  hp = strstr (readline, "camz:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      camz = atof (hilf);
	      camposspecified = YES;
	    }			/* if */
	  hp = strstr (readline, "Animate:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 8, (*ShellyArgs).animto, '\n');
	      (*ShellyArgs).anim = YES;
	    }			/* if */
	  hp = strstr (readline, "Steps:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 6, hilf, '\0');
	      (*ShellyArgs).steps = atoi (hilf);
	    }			/* if */
/*	  hp = strstr (readline, "Rand:");
	  if (hp != NULL)
	    {
	      mycopystr (hp + 5, hilf, '\0');
	      (*ShellyArgs).rnod = atof (hilf);
	    }*/			/* if */

	}			/* while */

      fprintf (stdout, "\n");
      if (fclose (fp) != 0)
	fprintf (stdout, "Error while closing file: %s!\n", fin);
    }
  else
    {
      fprintf (stdout, "Could not open datafile: %s!\n", fin);
      exit (5);
    }
}				/* ReadInfile */


/********************************************/
/* triangulate:                             */
/*                                          */
/********************************************/
int
triangulate (struct punkt *list1, struct punkt *list2, struct triangle **result)
{
  struct punkt *p11, *p12, *p21, *p22;
  struct triangle *tp, *tp2, *anker = NULL;
  int ende = 1;

  if ((list1 == NULL) || (list2 == NULL))
    return (5);

  p11 = list1;
  p21 = list2;

  while (ende)
    {
      p12 = p11->next;
      p22 = p21->next;


      if ((p11->next == NULL) && (p21->next == NULL))
	ende = 0;
      else if (p11->next == NULL)
	{
	  tp = calloc (1, sizeof (struct triangle));
	  if (tp != NULL)
	    {
	      tp->p1 = p21;
	      tp->p2 = p22;
	      tp->p3 = p11;
	      if (anker == NULL)
		anker = tp;
	      else
		tp2->next = tp;
	      tp2 = tp;
	    }
	  p21 = p22;
	}
      else if (p21->next == NULL)
	{
	  tp = calloc (1, sizeof (struct triangle));
	  if (tp != NULL)
	    {
	      tp->p1 = p11;
	      tp->p2 = p21;
	      tp->p3 = p12;
	      if (anker == NULL)
		anker = tp;
	      else
		tp2->next = tp;
	      tp2 = tp;
	    }
	  p11 = p12;
	}
      else if (p11->s == p21->s)
	{
	  if (fabs (p11->s - p12->s) < fabs (p21->s - p22->s))
	    {
	      tp = calloc (1, sizeof (struct triangle));
	      if (tp != NULL)
		{
		  tp->p1 = p11;
		  tp->p2 = p21;
		  tp->p3 = p12;
		  if (anker == NULL)
		    anker = tp;
		  else
		    tp2->next = tp;
		  tp2 = tp;
		}
	      p11 = p12;
	    }
	  else
	    {
	      tp = calloc (1, sizeof (struct triangle));
	      if (tp != NULL)
		{
		  tp->p1 = p21;
		  tp->p2 = p22;
		  tp->p3 = p11;
		  if (anker == NULL)
		    anker = tp;
		  else
		    tp2->next = tp;
		  tp2 = tp;
		}
	      p21 = p22;
	    }
	}
      else if (p11->s < p21->s)
	{
	  tp = calloc (1, sizeof (struct triangle));
	  if (tp != NULL)
	    {
	      tp->p1 = p11;
	      tp->p2 = p21;
	      tp->p3 = p12;
	      if (anker == NULL)
		anker = tp;
	      else
		tp2->next = tp;
	      tp2 = tp;
	    }
	  p11 = p12;

	}
      else if (p11->s > p21->s)
	{
	  tp = calloc (1, sizeof (struct triangle));
	  if (tp != NULL)
	    {
	      tp->p1 = p21;
	      tp->p2 = p22;
	      tp->p3 = p11;
	      if (anker == NULL)
		anker = tp;
	      else
		tp2->next = tp;
	      tp2 = tp;
	    }
	  p21 = p22;

	}			/* big if-cascade ;) */

    }				/* while */

  *result = anker;
  return (0);

}				/* triangulate */


/********************************************/
/* writepovheader:                          */
/*                                          */
/********************************************/
void
writepovheader (FILE * fp)
{

  fprintf (fp, "/* POV-Scenefile generated by Shelly1.6           */\n");
  fprintf (fp, "/* by RANDi: (rschultz@informatik.uni-rostock.de) */\n");

  fprintf (fp, "#declare te = pigment { color red 1 green 0 blue 0 }\n");

  fprintf (fp, "camera\n{\n location  <%.1f,%.1f,%.1f>\n", camx, camy, camz);
  fprintf (fp, " look_at <0, 0, 0>\n}\n");

  fprintf (fp, "object\n{\n light_source {\n");
  fprintf (fp, "  <250, 500, 500> color red 1 green 1 blue 1\n }\n");
  fprintf (fp, "}\n");
  fprintf (fp, "object\n{\n light_source {\n");
  fprintf (fp, "  <-250, 500, 500> color red 1 green 1 blue 1\n }\n");
  fprintf (fp, "}\n");

}				/* writepovheader */


/********************************************/
/* writepovtriangles:                       */
/********************************************/
int
writepovtriangles (FILE * fp, struct punkt *anker1, struct punkt *anker2, int mode)
{
  struct punkt *p11, *p12, *p21, *p22;
  struct triangle *tp, *tp1, *tp2;

  if ((anker1 == NULL) || (anker2 == NULL))
    return (0);

  if (mode == 1)
    {
      triangulate (anker1, anker2, &tp);

      tp1 = tp;
      while (tp1 != NULL)
	{

	  fprintf (fp, "triangle{<%.2f,%.2f,%.2f>,<%.2f,%.2f,%.2f>,<%.2f,%.2f,%.2f>\n", tp1->p1->x, tp1->p1->y, tp1->p1->z, tp1->p2->x, tp1->p2->y, tp1->p2->z, tp1->p3->x, tp1->p3->y, tp1->p3->z);
	  fprintf (fp, "pigment{te}}\n");

	  tp2 = tp1->next;

	  free (tp1);

	  tp1 = tp2;
	}			/* while */

      return (0);
    };


  p11 = anker1;
  p12 = p11;
  p21 = anker2;
  p22 = p21;

  do
    {
      p11 = p12;
      p21 = p22;
      p12 = (*p11).next;
      p22 = (*p21).next;

      if ((p12 != NULL) && (p22 != NULL))
	{
	  fprintf (fp, "triangle{<%.2f,%.2f,%.2f>,<%.2f,%.2f,%.2f>,<%.2f,%.2f,%.2f>\n", (*p11).x, (*p11).y, (*p11).z, (*p12).x, (*p12).y, (*p12).z, (*p21).x, (*p21).y, (*p21).z);
	  fprintf (fp, "pigment{te}}\n");

	  fprintf (fp, "triangle{<%.2f,%.2f,%.2f>,<%.2f,%.2f,%.2f>,<%.2f,%.2f,%.2f>\n", (*p21).x, (*p21).y, (*p21).z, (*p12).x, (*p12).y, (*p12).z, (*p22).x, (*p22).y, (*p22).z);
	  fprintf (fp, "pigment{te}}\n");
	}
    }
  while ((p12 != NULL) && (p22 != NULL));

  return (0);
}				/* writepovtriangles */

/********************************************/
/* writepovpatches:                         */
/* or: How to walk with 16 Pointer through  */
/*     4 Lists ;)                           */
/********************************************/
int
writepovpatches (FILE * fp, struct punkt *anker1, struct punkt *anker2, struct punkt *anker3, struct punkt *anker4)
{
  struct punkt *p[16];
  int i = 0;


  if ((anker1 == NULL) || (anker2 == NULL) || (anker3 == NULL) || (anker4 == NULL))
    return (0);

  p[0] = anker1;
  p[1] = anker2;
  p[2] = anker3;
  p[3] = anker4;

  do
    {
      if (p[0] && p[1] && p[2] && p[3])
	for (i = 0; i < 4; i++)
	  p[i + 4] = p[i]->next;

      if (p[4] && p[5] && p[6] && p[7])
	for (i = 0; i < 4; i++)
	  p[i + 8] = p[i + 4]->next;

      if (p[8] && p[9] && p[10] && p[11])
	for (i = 0; i < 4; i++)
	  p[i + 12] = p[i + 8]->next;

     if(p[0] && p[1] && p[2] && p[3] && p[4] && p[5] && p[6] && p[7] && p[8]&& p[9] && p[10] && p[11] && p[12] && p[13] && p[14] && p[15])
     {
      fprintf (fp, "bicubic_patch{type 1 flatness 0 u_steps 3 v_steps 3\n");
      fprintf (fp, "<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>\n", p[0]->x, p[0]->y, p[0]->z, p[1]->x, p[1]->y, p[1]->z, p[2]->x, p[2]->y, p[2]->z, p[3]->x, p[3]->y,p[3]->z);
      fprintf (fp, "<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,\n", p[4]->x, p[4]->y, p[4]->z, p[5]->x, p[5]->y, p[5]->z, p[6]->x, p[6]->y, p[6]->z, p[7]->x, p[7]->y, p[7]->z);
      fprintf (fp, "<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,\n", p[8]->x, p[8]->y, p[8]->z, p[9]->x, p[9]->y, p[9]->z, p[10]->x, p[10]->y, p[10]->z, p[11]->x, p[11]->y, p[11]->z);
      fprintf (fp, "<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>,<%.4f,%.4f,%.4f>\n", p[12]->x, p[12]->y, p[12]->z, p[13]->x, p[13]->y, p[13]->z, p[14]->x, p[14]->y, p[14]->z, p[15]->x, p[15]->y, p[15]->z);
      fprintf (fp, "pigment{te}}\n");
     } /* if */
      if (p[12] && p[13] && p[14] && p[15])
	for (i = 0; i < 4; i++)
	  p[i] = p[i + 12];

    }
  while (p[0] && p[1] && p[2] && p[3] && p[4] && p[5] && p[6] && p[7] && p[8] && p[9] && p[10] && p[11] && p[12] && p[13] && p[14] && p[15]);

  return (0);
}				/* writepovpatches */


/********************************************/
/* writepovfinish:                          */
/********************************************/
void
writepovfinish (FILE * fptemp, FILE * fp)
{

  filecopy (fptemp, fp);

}


/********************************************/
/* new_pnode:                               */
/*                                          */
/********************************************/
void
new_pnode (struct pmerk **anker, struct punkt *p1)
{
  struct pmerk *mp;

  if ((mp = calloc (1, sizeof (struct pmerk))) != NULL)
    {
      mp->x = p1->x;
      mp->y = p1->y;
      mp->z = p1->z;
      mp->id = gl.countp;
      mp->next = *anker;
      *anker = mp;
    }
  else
    return;

}				/* new_pnode */

/********************************************/
/* free_pmerk_list:                         */
/*                                          */
/********************************************/
void
free_pmerk_list (struct pmerk *anker)
{
  struct pmerk *mp1 = anker, *mp2;

  while (mp1 != NULL)
    {
      mp2 = mp1->next;
      free (mp1);
      mp1 = mp2;
    }				/* while */

}				/* free_pmerk_list */

/********************************************/
/* is_p_in:                                 */
/*                                          */
/********************************************/
int
is_p_in (struct pmerk *anker, struct punkt *p)
{
  struct pmerk *mp1 = anker, *mp2;

  while (mp1 != NULL)
    {
      if (mp1->x == p->x)
	if (mp1->y == p->y)
	  if (mp1->z == p->z)
	    return (mp1->id);

      mp2 = mp1->next;
      mp1 = mp2;
    }
  return (0);
}				/* is_p_in */

/********************************************/
/* new_lnode:                               */
/*                                          */
/********************************************/
void
new_lnode (struct lmerk **anker, int id1, int id2)
{
  struct lmerk *lp, *lp1 = *anker, *lp2;

  if (lp1 == NULL)
    {
      if ((lp = calloc (1, sizeof (struct lmerk))) != NULL)
	{
	  lp->id1 = id1;
	  lp->id2 = id2;
	  lp->id = gl.countl;
	  *anker = lp;
	}
      else
	return;
    }

  while (lp1 != NULL)
    {
      lp2 = lp1->next;
      if (lp2 == NULL)
	if ((lp = calloc (1, sizeof (struct lmerk))) != NULL)
	  {
	    lp->id1 = id1;
	    lp->id2 = id2;
	    lp->id = gl.countl;
	    lp1->next = lp;
	  }
	else
	  return;

      lp1 = lp2;
    }

}				/* new_lnode */

/********************************************/
/* app_llist:                               */
/*                                          */
/********************************************/
void
app_llist (struct lmerk *anker, struct lmerk *list)
{
  struct lmerk *lp1 = anker, *lp2;

  while (lp1 != NULL)
    {
      lp2 = lp1->next;
      if (lp2 == NULL)
	lp1->next = list;
      lp1 = lp2;
    }
  return;
}				/* app_llist */

/********************************************/
/* free_lmerk_list:                         */
/*                                          */
/********************************************/
void
free_lmerk_list (struct lmerk *anker)
{
  struct lmerk *mp1 = anker, *mp2;

  while (mp1 != NULL)
    {
      mp2 = mp1->next;
      free (mp1);
      mp1 = mp2;
    }				/* while */

}				/* free_lmerk_list */

/********************************************/
/* is_l_in:                                 */
/*                                          */
/********************************************/
int
is_l_in (struct lmerk *anker, int id1, int id2)
{
  struct lmerk *mp1 = anker, *mp2;

  while (mp1 != NULL)
    {
      if (mp1->id1 == id1)
	if (mp1->id2 == id2)
	  return (mp1->id);

      if (mp1->id2 == id1)
	if (mp1->id1 == id2)
	  return (mp1->id);

      mp2 = mp1->next;
      mp1 = mp2;
    }
  return (0);
}				/* is_l_in */

/********************************************/
/* new_tnode:                               */
/*                                          */
/********************************************/
void
new_tnode (struct tmerk **anker, int id1, int id2, int id3, int pid1, int pid2, int pid3)
{
  struct tmerk *tp, *tp1 = *anker, *tp2;
  if (tp1 == NULL)
    if ((tp = calloc (1, sizeof (struct tmerk))) != NULL)
      {
	tp->lid1 = id1;
	tp->lid2 = id2;
	tp->lid3 = id3;
	tp->pid1 = pid1;
	tp->pid2 = pid2;
	tp->pid3 = pid3;
	
	*anker = tp;
      }
    else
      return;
  while (tp1 != NULL)
    {
      tp2 = tp1->next;
      if (tp2 == NULL)
	if ((tp = calloc (1, sizeof (struct tmerk))) != NULL)
	  {
	    tp->lid1 = id1;
	    tp->lid2 = id2;
	    tp->lid3 = id3;
	    tp->pid1 = pid1;
	    tp->pid2 = pid2;
	    tp->pid3 = pid3;
            tp1->next = tp;
	  }
	else
	  return;
      tp1 = tp2;
    }

}				/* new_tnode */

/********************************************/
/* free_tmerk_list:                         */
/*                                          */
/********************************************/
void
free_tmerk_list (struct tmerk *anker)
{
  struct tmerk *tp1 = anker, *tp2;

  while (tp1 != NULL)
    {
      tp2 = tp1->next;
      free (tp1);
      tp1 = tp2;
    }				/* while */

}				/* free_tmerk_list */



/********************************************/
/* writerplline:                            */
/********************************************/
int
writerplline (FILE * fp, struct punkt *anker1, struct punkt *anker2, int mode)
{
  struct punkt *p11, *p12;
  struct triangle *tp1, *tp2, *tp;

  static struct pmerk *plist1 = NULL;
  struct pmerk *plist2 = NULL;

  int pid1, pid2, pid3;


  if (mode == 2)
    {
      free_pmerk_list (plist1);
      return (0);
    };

  if (mode == 1)
    {
      triangulate (anker1, anker2, &tp);

      tp1 = tp;
      while (tp1 != NULL)
	{
	  if (((pid1 = is_p_in (plist2, tp1->p1)) == 0) && ((pid1 = is_p_in (plist1, tp1->p1)) == 0))
	    {
	      fprintf (fp, "%.3f %.3f %.3f\n", tp1->p1->x/10.0, tp1->p1->y/10.0, tp1->p1->z/10.0);
	      new_pnode (&plist2, tp1->p1);
	      pid1 = gl.countp;
	      gl.countp++;
	    }

	  if (((pid2 = is_p_in (plist2, tp1->p2)) == 0) && ((pid2 = is_p_in (plist1, tp1->p2)) == 0))
	    {
	      fprintf (fp, "%.3f %.3f %.3f\n", tp1->p2->x/10.0, tp1->p2->y/10.0, tp1->p2->z/10.0);
	      new_pnode (&plist2, tp1->p2);
	      pid2 = gl.countp;
	      gl.countp++;
	    }

	  if (((pid3 = is_p_in (plist2, tp1->p3)) == 0) && ((pid3 = is_p_in (plist1, tp1->p3)) == 0))
	    {
	      fprintf (fp, "%.3f %.3f %.3f\n", tp1->p3->x/10.0, tp1->p3->y/10.0, tp1->p3->z/10.0);
	      new_pnode (&plist2, tp1->p3);
	      pid3 = gl.countp;
	      gl.countp++;
	    }


	  new_tnode (&bigtlist, 0, 0, 0, pid1, pid2, pid3);
	  gl.countt++;

	  tp2 = tp1->next;
	  free (tp1);
	  tp1 = tp2;
	}			/* while */


      free_pmerk_list (plist1);
      plist1 = plist2;

      return (0);
    };

  if (anker1 == NULL)
    return (0);

  p11 = anker1;
  p12 = p11;

  gl.countw++;			/* calc number of lines created */

  fprintf (fp, "%.3f  %.3f  %.3f\n", (*p11).x / 100.0, (*p11).y / 100.0, (*p11).z / 100.0);

  do
    {
      p11 = p12;
      p12 = (*p11).next;

      if (p12 != NULL)
	{
	  fprintf (fp, "%.3f  %.3f  %.3f\n", (*p12).x / 100.0, (*p12).y / 100.0, (*p12).z / 100.0);
	}
    }
  while (p12 != NULL);

  return (0);
}				/* writerplline */


/********************************************/
/* writerplfinish:                          */
/*                                          */
/********************************************/
void
writerplfinish (FILE * fptemp, FILE * fp, int mode)
{
  int faces = 0;
  struct tmerk *tp1 = bigtlist, *tp2;

  filecopy (fptemp, fp);

  if (mode == 1)
   {
    fprintf (fp, "%d ( Points\n", gl.countp);	/* Number of points */
 
    while (tp1 != NULL)
    {
     fprintf (fp, "%d %d %d\n", tp1->pid1, tp1->pid2, tp1->pid3);
     tp2 = tp1->next;
     free (tp1);
     tp1 = tp2;
    }

    faces = gl.countt;
    fprintf (fp, "%d ( Faces\n", faces);

    fprintf (fp, "wFT_PHONG\n255 255 255 0\n");	/* type RGBA */
    fprintf (fp, "\"Shell\"\n0\n");	/* Name */
    fprintf (fp, "\"CEND\"\n");
    fprintf (fp, "C_TRISET DROP\n");


    return;
   }
   
  if (mode == 0)
    {
      fprintf (fp, "%d\n", gl.counth);	/* height width of the mesh */
      fprintf (fp, "%d\n", gl.countw);

      fprintf (fp, "3\n0\n255 255 255 0\n");	/* RGBA */
      fprintf (fp, "\"Shell\"\n0\n");	/* Name */
      fprintf (fp, "\"CEND\"\n");
      fprintf (fp, "C_MESH DROP\n");
    }				/* if */
}				/* writerplfinish */

/********************************************/
/* writex3dheader:                          */
/*                                          */
/********************************************/
void
writex3dheader (FILE * fp)
{

  fprintf (fp, "2\n");		/* Number of Colors */
  fprintf (fp, "0  0 0 0\n");	/* Color 0 Black */
  fprintf (fp, "1  255 0 0\n");	/* Color 1 Red */

}				/* writex3dheader */


/********************************************/
/* writex3dline:                            */
/********************************************/
int
writex3dline (FILE * fp, struct punkt *anker1, struct punkt *anker2, int mode)
{
  struct punkt *p11, *p12;
  struct triangle *tp, *tp1, *tp2;

  static struct pmerk *plist1 = NULL;
  struct pmerk *plist2 = NULL;
  static struct lmerk *llist1 = NULL;
  struct lmerk *llist2 = NULL;

  int pid1, pid2, pid3;
  int lid1, lid2, lid3;

  if (mode == 2)
    {
      free_pmerk_list (plist1);
      return (0);
    };

  if (mode == 1)
    {
      triangulate (anker1, anker2, &tp);

      tp1 = tp;
      while (tp1 != NULL)
	{
	  if (((pid1 = is_p_in (plist2, tp1->p1)) == 0) && ((pid1 = is_p_in (plist1, tp1->p1)) == 0))
	    {
	      fprintf (fp, "%d %.3f %.3f %.3f\n", gl.countp, tp1->p1->x, tp1->p1->y, tp1->p1->z);
	      new_pnode (&plist2, tp1->p1);
	      pid1 = gl.countp;
	      gl.countp++;
	    }

	  if (((pid2 = is_p_in (plist2, tp1->p2)) == 0) && ((pid2 = is_p_in (plist1, tp1->p2)) == 0))
	    {
	      fprintf (fp, "%d %.3f %.3f %.3f\n", gl.countp, tp1->p2->x, tp1->p2->y, tp1->p2->z);
	      new_pnode (&plist2, tp1->p2);
	      pid2 = gl.countp;
	      gl.countp++;
	    }

	  if (((pid3 = is_p_in (plist2, tp1->p3)) == 0) && ((pid3 = is_p_in (plist1, tp1->p3)) == 0))
	    {
	      fprintf (fp, "%d %.3f %.3f %.3f\n", gl.countp, tp1->p3->x, tp1->p3->y, tp1->p3->z);
	      new_pnode (&plist2, tp1->p3);
	      pid3 = gl.countp;
	      gl.countp++;
	    }

	  if (((lid1 = is_l_in (llist2, pid1, pid2)) == 0) && ((lid1 = is_l_in (llist1, pid1, pid2)) == 0))
	    {
	      new_lnode (&llist2, pid1, pid2);
	      lid1 = gl.countl;
	      gl.countl++;
	    }

	  if (((lid2 = is_l_in (llist2, pid2, pid3)) == 0) && ((lid2 = is_l_in (llist1, pid2, pid3)) == 0))
	    {
	      new_lnode (&llist2, pid2, pid3);
	      lid2 = gl.countl;
	      gl.countl++;
	    }

	  if (((lid3 = is_l_in (llist2, pid3, pid1)) == 0) && ((lid3 = is_l_in (llist1, pid3, pid1)) == 0))
	    {
	      new_lnode (&llist2, pid3, pid1);
	      lid3 = gl.countl;
	      gl.countl++;
	    }

	  new_tnode (&bigtlist, lid1, lid2, lid3, 0, 0, 0);
	  gl.countt++;

	  tp2 = tp1->next;
	  free (tp1);
	  tp1 = tp2;
	}			/* while */


      free_pmerk_list (plist1);
      plist1 = plist2;
      if (gl.calledwrtx3dl == 0)
	bigllist = llist2;
      else
	app_llist (llist1, llist2);
      llist1 = llist2;

      gl.calledwrtx3dl = 1;
      return (0);
    };



  if (anker1 == NULL)
    return (0);

  if (gl.calledwrtx3dl == 0)
    {
      gl.calledwrtx3dl = 1;
      gl.count = -1;
      gl.countw = 0;
    }
  gl.countw++;

  p11 = anker1;
  p12 = p11;

  fprintf (fp, "%d  %.3f %.3f %.3f\n", ++gl.count, (*p11).x, (*p11).y, (*p11).z);

  do
    {
      p11 = p12;
      p12 = (*p11).next;

      if (p12 != NULL)
	{
	  fprintf (fp, "%d  %.3f %.3f %.3f\n", ++gl.count, (*p12).x, (*p12).y, (*p12).z);
	}
    }
  while (p12 != NULL);

  return (0);
}				/* writex3dline */

/********************************************/
/* writex3dfinish:                          */
/*                                          */
/********************************************/
void
writex3dfinish (FILE * fptemp, FILE * fp, int mode)
{
  int edges = 0, faces = 0, edgecount = 0, facecount = 0, a = 0, b = 0,
    r = 0, r2 = 0, r3 = 0;
  int point1 = 0, point2 = 0, edge1 = 0, edge2 = 0, edge3 = 0;
  struct tmerk *tp1 = bigtlist, *tp2;
  struct lmerk *lp1 = bigllist, *lp2;

  fprintf (stdout, "\nWriting additional X3D-Data ...");

  fprintf (fp, "%d\n", gl.countp);	/* Number of points */

  filecopy (fptemp, fp);

  if (mode == 1)
    {
      edges = gl.countl;
      fprintf (fp, "%d\n", edges);

      while (lp1 != NULL)
	{
	  fprintf (fp, "%d 1 %d %d\n", edgecount++, lp1->id1, lp1->id2);
	  lp2 = lp1->next;
	  free (lp1);
	  lp1 = lp2;
	}

      faces = gl.countt;
      fprintf (fp, "%d\n", faces);

      while (tp1 != NULL)
	{
	  fprintf (fp, "%d 1 3 %d %d %d\n", facecount++, tp1->lid1, tp1->lid2, tp1->lid3);
	  tp2 = tp1->next;
	  free (tp1);
	  tp1 = tp2;
	}

      return;
    }

  edges = (gl.counth * (gl.countw - 1) + gl.countw * (gl.counth - 1) + (gl.counth - 1) * (gl.countw - 1));
  /* number of edges! */
  fprintf (fp, "%d\n", edges);


  /* write horizontal edges */
  for (a = 0; a < gl.counth; a++)
    {
      r = a;
      for (b = 0; b < (gl.countw - 1); b++)
	{
	  point1 = r;
	  point2 = r + gl.counth;
	  fprintf (fp, "%d  1   %d %d\n", edgecount++, point1, point2);
	  r = r + gl.counth;
	}
    }
  /* write vertical edges */
  for (a = 0; a < gl.countw; a++)
    {
      r = a * gl.counth;
      for (b = 0; b < (gl.counth - 1); b++)
	{
	  point1 = r + b;
	  point2 = r + b + 1;
	  fprintf (fp, "%d  1   %d %d\n", edgecount++, point1, point2);
	}
    }
  /* write diagonal edges */
  for (a = 0; a < gl.counth - 1; a++)
    {
      r = a;
      for (b = 0; b < (gl.countw - 1); b++)
	{
	  point1 = r;
	  point2 = r + gl.counth + 1;
	  fprintf (fp, "%d  1   %d %d\n", edgecount++, point1, point2);
	  r = r + gl.counth;
	}
    }
  /* ohhh boy, THIS (^^^) took me time ! :), now to that faces! */

  /* write faces */

  faces = 2 * (gl.countw - 1) * (gl.counth - 1);

  fprintf (fp, "%d\n", faces);	/* number of faces/polygons */

  r = 0;
  r3 = gl.countw * (gl.counth - 1) + gl.counth * (gl.countw - 1);

  for (a = 0; a < gl.counth - 1; a++)
    {
      r2 = gl.counth * (gl.countw - 1) + a;
      for (b = 0; b < gl.countw - 1; b++)
	{
	  edge1 = r + gl.countw - 1;
	  edge2 = r2;
	  edge3 = r3;
	  fprintf (fp, "%d  1  3  %d %d %d\n", facecount++, edge1, edge2, edge3);

	  edge1 = r;
	  edge2 = r2 + gl.counth - 1;
	  edge3 = r3;
	  fprintf (fp, "%d  1  3  %d %d %d\n", facecount++, edge1, edge2, edge3);

	  r++;
	  r2 = r2 + gl.counth - 1;
	  r3++;
	}
    }


}				/* writex3dfinish */


/********************************************/
/* writet3dheader:                          */
/*                                          */
/********************************************/
void
writet3dheader (FILE * fp)
{
  char percent = '%';

  fprintf (fp, " %c T3Dlib-data-file generated by Shelly1.6\n", percent);
  fprintf (fp, " %c the ShellShapeGenerator by RANDi :    \n", percent);
  fprintf (fp, " %c (rschultz@informatik.uni-rostock.de)\n", percent);

  fprintf (fp, "OBJ Begin \"Hierarchy 1\"\n");
  fprintf (fp, " DESC Begin \"Object 1 at level 1 of hierarchy 1\"\n");
  fprintf (fp, "  NAME \"SHELL\"\n");
  fprintf (fp, "  SHAP Shape = 2\n");
  fprintf (fp, "  SHAP Lamp  = 0\n");
  fprintf (fp, "  POSI X=0 Y=0 Z=0\n");
  fprintf (fp, "  AXIS XAxis X=1 Y=0 Z=0\n");
  fprintf (fp, "  AXIS YAxis X=0 Y=1 Z=0\n");
  fprintf (fp, "  AXIS ZAxis X=0 Y=0 Z=1\n");
  fprintf (fp, "  SIZE X=32 Y=32 Z=32\n");

}				/* writet3dheader */


/********************************************/
/* writet3dline:                            */
/********************************************/
int
writet3dline (FILE * fp, struct punkt *anker1, struct punkt *anker2, int mode)
{
  struct punkt *p11, *p12;
  struct triangle *tp, *tp1, *tp2;

  static struct pmerk *plist1 = NULL;
  struct pmerk *plist2 = NULL;
  static struct lmerk *llist1 = NULL;
  struct lmerk *llist2 = NULL;


  int pid1, pid2, pid3;
  int lid1, lid2, lid3;

  if ((anker1 == NULL) && (anker2 == NULL))
    return (5);

  if (mode == 2)
    {
      free_pmerk_list (plist1);
      return (0);
    };

  if (mode == 1)
    {
      triangulate (anker1, anker2, &tp);

      tp1 = tp;
      while (tp1 != NULL)
	{
	  if (((pid1 = is_p_in (plist2, tp1->p1)) == 0) && ((pid1 = is_p_in (plist1, tp1->p1)) == 0))
	    {
	      fprintf (fp, "  PNTS Point[%d] X=%.3f Y=%.3f Z=%.3f\n", gl.countp, tp1->p1->x, tp1->p1->y, tp1->p1->z);
	      new_pnode (&plist2, tp1->p1);
	      pid1 = gl.countp;
	      gl.countp++;
	    }

	  if (((pid2 = is_p_in (plist2, tp1->p2)) == 0) && ((pid2 = is_p_in (plist1, tp1->p2)) == 0))
	    {
	      fprintf (fp, "  PNTS Point[%d] X=%.3f Y=%.3f Z=%.3f\n", gl.countp, tp1->p2->x, tp1->p2->y, tp1->p2->z);
	      new_pnode (&plist2, tp1->p2);
	      pid2 = gl.countp;
	      gl.countp++;
	    }

	  if (((pid3 = is_p_in (plist2, tp1->p3)) == 0) && ((pid3 = is_p_in (plist1, tp1->p3)) == 0))
	    {
	      fprintf (fp, "  PNTS Point[%d] X=%.3f Y=%.3f Z=%.3f\n", gl.countp, tp1->p3->x, tp1->p3->y, tp1->p3->z);
	      new_pnode (&plist2, tp1->p3);
	      pid3 = gl.countp;
	      gl.countp++;
	    }

	  if (((lid1 = is_l_in (llist2, pid1, pid2)) == 0) && ((lid1 = is_l_in (llist1, pid1, pid2)) == 0))
	    {
	      new_lnode (&llist2, pid1, pid2);
	      lid1 = gl.countl;
	      gl.countl++;
	    }

	  if (((lid2 = is_l_in (llist2, pid2, pid3)) == 0) && ((lid2 = is_l_in (llist1, pid2, pid3)) == 0))
	    {
	      new_lnode (&llist2, pid2, pid3);
	      lid2 = gl.countl;
	      gl.countl++;
	    }

	  if (((lid3 = is_l_in (llist2, pid3, pid1)) == 0) && ((lid3 = is_l_in (llist1, pid3, pid1)) == 0))
	    {
	      new_lnode (&llist2, pid3, pid1);
	      lid3 = gl.countl;
	      gl.countl++;
	    }

	  new_tnode (&bigtlist, lid1, lid2, lid3,0,0,0);
	  gl.countt++;

	  tp2 = tp1->next;
	  free (tp1);
	  tp1 = tp2;
	}			/* while */


      free_pmerk_list (plist1);
      plist1 = plist2;

      if (gl.calledwrtt3dl == 0)
	bigllist = llist2;
      else
	app_llist (llist1, llist2);
      llist1 = llist2;

      gl.calledwrtt3dl = 1;
      return (0);
    };


  if (anker1 == NULL)
    return (0);

  if (gl.calledwrtt3dl == 0)
    {
      gl.calledwrtt3dl = 1;
      gl.count = -1;
      gl.countw = 0;
    }
  gl.countw++;

  p11 = anker1;
  p12 = p11;

  fprintf (fp, "  PNTS Point[%d] X=%.3f Y=%.3f Z=%.3f\n", ++gl.count, (*p11).x, (*p11).y, (*p11).z);

  do
    {
      p11 = p12;
      p12 = (*p11).next;

      if (p12 != NULL)
	{
	  fprintf (fp, "  PNTS Point[%d] X=%.3f Y=%.3f Z=%.3f\n", ++gl.count, (*p12).x, (*p12).y, (*p12).z);
	}
    }
  while (p12 != NULL);

  return (0);
}				/* writet3dline */


/********************************************/
/* writet3dfinish:                          */
/*                                          */
/********************************************/
void
writet3dfinish (FILE * fptemp, FILE * fp, int mode)
{
  int edges = 0, faces = 0, edgecount = 0, facecount = 0, a = 0, b = 0,
    r = 0, r2 = 0, r3 = 0;
  int point1 = 0, point2 = 0, edge1 = 0, edge2 = 0, edge3 = 0;
  struct tmerk *tp1 = bigtlist, *tp2;
  struct lmerk *lp1 = bigllist, *lp2;


  fprintf (stdout, "\nWriting additional T3Dlib-Data ...");

  fprintf (fp, "  PNTS PCount %d\n", gl.countp);

  filecopy (fptemp, fp);

  if (mode == 1)
    {
      edges = gl.countl;
      fprintf (fp, "  EDGE ECount %d\n", edges);

      while (lp1 != NULL)
	{
	  fprintf (fp, "  EDGE Edge[%d] %d %d\n", edgecount++, lp1->id1, lp1->id2);
	  lp2 = lp1->next;
	  free (lp1);
	  lp1 = lp2;
	}

      faces = gl.countt;
      fprintf (fp, "  FACE TCount %d\n", faces);

      while (tp1 != NULL)
	{
	  fprintf (fp, "  FACE Connect[%d] %d %d %d\n", facecount++, tp1->lid1, tp1->lid2, tp1->lid3);
	  tp2 = tp1->next;
	  free (tp1);
	  tp1 = tp2;
	}

      return;
    }

  edges = (gl.counth * (gl.countw - 1) + gl.countw * (gl.counth - 1) + (gl.counth - 1) * (gl.countw - 1));
  /* number of edges! */
  fprintf (fp, "  EDGE ECount %d\n", edges);


  /* write horizontal edges */
  for (a = 0; a < gl.counth; a++)
    {
      r = a;
      for (b = 0; b < (gl.countw - 1); b++)
	{
	  point1 = r;
	  point2 = r + gl.counth;
	  fprintf (fp, "  EDGE Edge[%d] %d %d\n", edgecount++, point1, point2);
	  r = r + gl.counth;
	}
    }
  /* write vertical edges */
  for (a = 0; a < gl.countw; a++)
    {
      r = a * gl.counth;
      for (b = 0; b < (gl.counth - 1); b++)
	{
	  point1 = r + b;
	  point2 = r + b + 1;
	  fprintf (fp, "  EDGE Edge[%d] %d %d\n", edgecount++, point1, point2);
	}
    }
  /* write diagonal edges */
  for (a = 0; a < gl.counth - 1; a++)
    {
      r = a;
      for (b = 0; b < (gl.countw - 1); b++)
	{
	  point1 = r;
	  point2 = r + gl.counth + 1;
	  fprintf (fp, "  EDGE Edge[%d] %d %d\n", edgecount++, point1, point2);
	  r = r + gl.counth;
	}
    }
  /* ohhh boy, THIS (^^^) took me time ! :), now to that faces! */

  /* write faces */

  faces = 2 * (gl.countw - 1) * (gl.counth - 1);

  fprintf (fp, "  FACE TCount %d\n", faces);

  r = 0;
  r3 = gl.countw * (gl.counth - 1) + gl.counth * (gl.countw - 1);

  for (a = 0; a < gl.counth - 1; a++)
    {
      r2 = gl.counth * (gl.countw - 1) + a;
      for (b = 0; b < gl.countw - 1; b++)
	{
	  edge1 = r + gl.countw - 1;
	  edge2 = r2;
	  edge3 = r3;
	  fprintf (fp, "  FACE Connect[%d] %d %d %d\n", facecount++, edge1, edge2, edge3);

	  edge1 = r;
	  edge2 = r2 + gl.counth - 1;
	  edge3 = r3;
	  fprintf (fp, "  FACE Connect[%d] %d %d %d\n", facecount++, edge1, edge2, edge3);

	  r++;
	  r2 = r2 + gl.counth - 1;
	  r3++;
	}
    }

  fprintf (fp, "  COLR R=98 G=68 B=58\n");
  fprintf (fp, "  REFL R=0 G=0 B=0\n");
  fprintf (fp, "  TRAN R=0 G=0 B=0\n");
  fprintf (fp, "  SPC1 R=0 G=0 B=0\n");
  fprintf (fp, "  End DESC   \"Object 1 at level 1 of hierarchy 1\"\n");
  fprintf (fp, " TOBJ       \"Object 0 at level 1 of hierarchy 1\"\n");
  fprintf (fp, "End OBJ  \"Hierarchy 1\"\n");

}				/* writet3dfinish */


/********************************************/
/* writerawtriangles:                       */
/********************************************/
int
writerawtriangles (FILE * fp, struct punkt *anker1, struct punkt *anker2, int mode)
{
  struct punkt *p11, *p12, *p21, *p22;
  struct triangle *tp, *tp1, *tp2;

  if ((anker1 == NULL) || (anker2 == NULL))
    return (0);

  if (mode == 1)
    {
      triangulate (anker1, anker2, &tp);

      tp1 = tp;
      while (tp1 != NULL)
	{
	  fprintf (fp, "%.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f\n", tp1->p1->x, tp1->p1->y, tp1->p1->z, tp1->p2->x, tp1->p2->y, tp1->p2->z, tp1->p3->x, tp1->p3->y, tp1->p3->z);
	  tp2 = tp1->next;

	  free (tp1);

	  tp1 = tp2;
	}			/* while */

      return (0);
    };

  p11 = anker1;
  p12 = p11;
  p21 = anker2;
  p22 = p21;

  do
    {
      p11 = p12;
      p21 = p22;
      p12 = (*p11).next;
      p22 = (*p21).next;

      if ((p12 != NULL) && (p22 != NULL))
	{
	  fprintf (fp, "%.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f\n", (*p11).x, (*p11).y, (*p11).z, (*p12).x, (*p12).y, (*p12).z, (*p21).x, (*p21).y, (*p21).z);

	  fprintf (fp, "%.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f\n", (*p21).x, (*p21).y, (*p21).z, (*p12).x, (*p12).y, (*p12).z, (*p22).x, (*p22).y, (*p22).z);
	}
    }
  while ((p12 != NULL) && (p22 != NULL));

  return (0);
}				/* writerawtriangles */


/********************************************/
/*init: fills global structure              */
/********************************************/
void
init (struct ShellyArguments *ShellyArgs)
{
  gl.countw = 0, gl.counth = 0, gl.countp = 0, gl.countl = 0, gl.countt = 0;
  gl.calledwrtx3dl = 0;
  gl.count = 0;
  gl.calledwrtt3dl = 0;
  gl.N = (*ShellyArgs).N;
  gl.N2 = (*ShellyArgs).N2;
  gl.N3 = (*ShellyArgs).N3;
  gl.L = (*ShellyArgs).L;
  gl.L2 = (*ShellyArgs).L2;
  gl.L3 = (*ShellyArgs).L3;
  gl.W1 = (*ShellyArgs).W1 * pi / 180;
  gl.W2 = (*ShellyArgs).W2 * pi / 180;
  gl.W12 = (*ShellyArgs).W12 * pi / 180;
  gl.W22 = (*ShellyArgs).W22 * pi / 180;
  gl.W13 = (*ShellyArgs).W13 * pi / 180;
  gl.W23 = (*ShellyArgs).W23 * pi / 180;
  gl.P = (*ShellyArgs).P * pi / 180;
  gl.P2 = (*ShellyArgs).P2 * pi / 180;
  gl.P3 = (*ShellyArgs).P3 * pi / 180;
  gl.Off2 = (*ShellyArgs).Off2 * pi / 180;
  gl.Off3 = (*ShellyArgs).Off3 * pi / 180;
  gl.biggest = 0;
  gl.alpha = (*ShellyArgs).alpha * pi / 180;
  gl.beta = (*ShellyArgs).beta * pi / 180;
  gl.phi = (*ShellyArgs).phi * pi / 180;
  gl.omega = (*ShellyArgs).omega * pi / 180;
  gl.my = (*ShellyArgs).my * pi / 180;

  gl.a = (*ShellyArgs).a;
  gl.b = (*ShellyArgs).b;
  gl.A = (*ShellyArgs).A;

  gl.Scale = (*ShellyArgs).Scale;

  gl.smin = (*ShellyArgs).smin * pi / 180;
  gl.smax = (*ShellyArgs).smax * pi / 180;
  gl.sd = (*ShellyArgs).sd * pi / 180;
  gl.Scans = (*ShellyArgs).Scans;
  gl.omin = (*ShellyArgs).omin * pi / 180;
  gl.omax = (*ShellyArgs).omax * pi / 180;
  gl.od = (*ShellyArgs).od * pi / 180;
  gl.Scano = (*ShellyArgs).Scano;

}				/* init */


/********************************************/
/*calcg:                                    */
/********************************************/
double
calcg (double O, double N)
{
  if (N == 0)
    return (0);
  else
    return ((2 * pi) / N * (O * N / (2 * pi) - round (O * N / (2 * pi))));

}				/* calcg */


/********************************************/
/*calck:                                    */
/********************************************/
double
calck (double S, double O)
{
  double N, N2, N3, L, L2, L3, W1, W2, W12, W13, W22, W23, P, P2, P3, Off2,
    Off3;
  double k = 0, g;


  N = gl.N;
  W1 = gl.W1;
  W2 = gl.W2;
  P = gl.P;
  L = gl.L;
  N2 = gl.N2;
  W12 = gl.W12;
  W22 = gl.W22;
  P2 = gl.P2;
  L2 = gl.L2;

  N3 = gl.N3;
  W13 = gl.W13;
  W23 = gl.W23;
  P3 = gl.P3;
  L3 = gl.L3;

  Off2 = gl.Off2;
  Off3 = gl.Off3;




  g = calcg (O, N);
  if (N != 0)
    k = L * exp (-(2 * (S - P) / W1) * (2 * (S - P) / W1)) * exp (-(2 * g / W2) * (2 * g / W2));

  g = calcg (O + Off2, N2);
  if (N2 != 0)
    k = k + (L2 * exp (-(2 * (S - P2) / W12) * (2 * (S - P2) / W12)) * exp (-(2 * g / W22) * (2 * g / W22)));

  g = calcg (O + Off3, N3);
  if (N3 != 0)
    k = k + (L3 * exp (-(2 * (S - P3) / W13) * (2 * (S - P3) / W13)) * exp (-(2 * g / W23) * (2 * g / W23)));

/*  if(gl.rnod!=0.0)
  {
  
  
   k = k + ((double) RAND_MAX/rand())*gl.rnod;
   printf("k= %f %f\n",k,gl.rnod);
  }*/

  return (k);
}				/* calck */


/********************************************/
/*innodule2:                                */
/********************************************/
double
innodule2 (double sc, double O, double S)
{
  double k1 = 0, k2 = 0;

  k1 = calck (S, O);

  k2 = calck (S + sc, O);

  return (fabs (k1) - fabs (k2));

}				/* innodule */


/********************************************/
/*innodule:                                 */
/********************************************/
double
innodule (double sc, double O)
{
  double N, N2, N3, L, L2, L3, W2, W22, W23, Off2, Off3;

  double k1 = 0, k2 = 0, g;

  N = gl.N;
  W2 = gl.W2;
  L = gl.L;
  N2 = gl.N2;
  W22 = gl.W22;
  L2 = gl.L2;
  N3 = gl.N3;
  W23 = gl.W23;
  L3 = gl.L3;
  Off2 = gl.Off2;
  Off3 = gl.Off3;


  g = calcg (O, N);
  if (N != 0)
    k1 = L * exp (-(2 * g / W2) * (2 * g / W2));
  g = calcg (O + Off2, N2);
  if (N2 != 0)
    k1 = k1 + (L2 * exp (-(2 * g / W22) * (2 * g / W22)));
  g = calcg (O + Off3, N3);
  if (N3 != 0)
    k1 = k1 + (L3 * exp (-(2 * g / W23) * (2 * g / W23)));

  g = calcg (O + sc, N);
  if (N != 0)
    k2 = L * exp (-(2 * g / W2) * (2 * g / W2));
  g = calcg (O + Off2 + sc, N2);
  if (N2 != 0)
    k2 = k2 + (L2 * exp (-(2 * g / W22) * (2 * g / W22)));
  g = calcg (O + Off3 + sc, N3);
  if (N3 != 0)
    k2 = k2 + (L3 * exp (-(2 * g / W23) * (2 * g / W23)));


  return (fabs (k1) - fabs (k2));

}				/* innodule */


/********************************************/
/*calccamera:                               */
/********************************************/
void
calccamera (double x, double y, double z)
{

  if ((fabs (x) > fabs (y)) && (fabs (x) > fabs (z)))
    if (fabs (x) > gl.biggest)
      {
	gl.biggest = fabs (x);
	camy = gl.biggest * 1.5;
	camz = gl.biggest * 1.5;
      }
  if ((fabs (y) > fabs (x)) && (fabs (y) > fabs (z)))
    if (fabs (y) > gl.biggest)
      {
	gl.biggest = fabs (y);
	camy = gl.biggest * 1.5;
	camz = gl.biggest * 1.5;
      }

  if ((fabs (z) > fabs (y)) && (fabs (z) > fabs (x)))
    if (fabs (z) > gl.biggest)
      {
	gl.biggest = fabs (z);
	camy = gl.biggest * 1.5;
	camz = gl.biggest * 1.5;
      }

}				/* calccamera */


/**********************************************/
/* CalcPoint:                                 */
/*                                            */
/**********************************************/
int
calcpoint (double O, double S, struct punkt **point)
{
  struct punkt *point1 = NULL;
  double x, y, z, R, Re, k;
  double a, b, A, beta, phi, omega, alpha, Scale, my;

  alpha = gl.alpha;
  beta = gl.beta;
  phi = gl.phi;
  omega = gl.omega;
  my = gl.my;

  a = gl.a;
  b = gl.b;
  A = gl.A;

  Scale = gl.Scale;

  gl.counth = 0;

  Re = pow ((pow (a, -2.0) * (cos (S) * cos (S)) + pow (b, -2.0) * (sin (S) * sin (S))), -0.5);

  k = calck (S, O);

  R = Re + k;

  /* alternatively: */
  /* R = pow((pow(a,2.0)*pow(cos(S),2.0)+pow(b,2.0)*pow(sin(S),2.0)),0.5); */

  x = Scale * (A * sin (beta) * cos (O) + R * cos (S + phi) * cos (O + omega) - R * sin (my) * sin (S + phi) * sin (O)) * exp (O * cot (alpha));
  y = Scale * (-1 * A * sin (beta) * sin (O) - R * cos (S + phi) * sin (O + omega) - R * sin (my) * sin (S + phi) * cos (O)) * exp (O * cot (alpha));
  z = Scale * (-1 * A * cos (beta) + R * sin (S + phi) * cos (my)) * exp (O * cot (alpha));
  if (camposspecified == NO)
    calccamera (x, y, z);
  if ((point1 = calloc (1, sizeof (struct punkt))) == NULL)
    {
      fprintf (stdout, "Shelly: NOT ENOUGH MEMORY!!! ... exiting ... !\n");
      return (5);
    }				/* if */

  (*point1).x = x;
  (*point1).y = y;
  (*point1).z = z;
  (*point1).s = S;

  *point = point1;

  return (0);

}				/* calcpoint */


/**********************************************/
/* calcnews:                                  */
/*             SSIA :)                        */
/**********************************************/
void
calcnews (struct ShellyArguments *SA, double O, double Sold, double *newS)
{
  double S = Sold, sd;
  int mittenodul = NO;
  sd = SA->sd * pi / 180;

  S += SA->Scans;


  while ((fabs (innodule2 (S - Sold, O, Sold)) < SA->heightdiffs) && (S - Sold < sd) && (mittenodul == NO))
    {
      S += SA->Scans;
      if ((innodule2 (SA->Scans, O, Sold) < 0) && (innodule2 (SA->Scans, O, S) > 0))
	{
	  mittenodul = YES;
	}
    }

  *newS = S;

  return;

}				/* calcnews */


/**********************************************/
/* CalcnodLine:                               */
/*                                            */
/**********************************************/
int
calcnodline (double O, struct ShellyArguments *ShellyArgs, struct punkt **list)
{
  struct punkt *list1 = NULL, *L1p1, *L1p2;
  double S, newS;
  double sc;
  double smin, smax, sd;

  smin = gl.smin;
  smax = gl.smax;
  sd = gl.sd;

  sc = gl.Scans;


  S = smin;

  calcpoint (O, smin, &list1);
  L1p1 = list1;
  L1p2 = list1;

  while (S <= smax)
    {
      calcnews (ShellyArgs, O, S, &newS);
      S = newS;

      calcpoint (O, S, &L1p1);
      L1p2->next = L1p1;
      L1p2 = L1p1;


    }				/* while */

  *list = list1;

  return (0);

}				/* calcnodline */


/**********************************************/
/* CalcLine:                                  */
/*                                            */
/**********************************************/
int
calcline (double O, struct ShellyArguments *ShellyArgs, struct punkt **list)
{
  struct punkt *list1 = NULL, *L1p1, *L1p2;
  double x, y, z, R, S, Re, k;
  double my;
  double smin, smax, sd, a, b, A, beta, phi, omega, alpha, Scale;

  smin = gl.smin;
  smax = gl.smax;
  sd = gl.sd;

  alpha = gl.alpha;
  beta = gl.beta;
  phi = gl.phi;
  omega = gl.omega;
  my = gl.my;

  a = gl.a;
  b = gl.b;
  A = gl.A;

  Scale = gl.Scale;

  gl.counth = 0;

  for (S = smin; (S + sd) < smax; S = S + sd)
    {
      gl.counth++;
      gl.countp++;

      Re = pow ((pow (a, -2.0) * (cos (S) * cos (S)) + pow (b, -2.0) * (sin (S) * sin (S))), -0.5);

      k = calck (S, O);

      R = Re + k;

      /* alternatively: */
      /* R = pow((pow(a,2.0)*pow(cos(S),2.0)+pow(b,2.0)*pow(sin(S),2.0)),0.5); */

      x = Scale * (A * sin (beta) * cos (O) + R * cos (S + phi) * cos (O + omega) - R * sin (my) * sin (S + phi) * sin (O)) * exp (O * cot (alpha));
      y = Scale * (-1 * A * sin (beta) * sin (O) - R * cos (S + phi) * sin (O + omega) - R * sin (my) * sin (S + phi) * cos (O)) * exp (O * cot (alpha));
      z = Scale * (-1 * A * cos (beta) + R * sin (S + phi) * cos (my)) * exp (O * cot (alpha));
      if (camposspecified == NO)
	calccamera (x, y, z);

      if ((L1p1 = calloc (1, sizeof (struct punkt))) == NULL)
	{
	  fprintf (stdout, "Shelly: NOT ENOUGH MEMORY!!! ... exiting ... !\n");
	  if (S != smin)
	    myfree (list1);
	  return (5);
	}			/* if */
      if (S == smin)
	list1 = L1p1;
      else
	(*L1p2).next = L1p1;
      L1p2 = L1p1;

      (*L1p1).x = x;
      (*L1p1).y = y;
      (*L1p1).z = z;
      (*L1p1).s = S;

    }				/* for */

  *list = list1;

  return (0);

}				/* calcline */


/**********************************************/
/* calcnewo:                                  */
/*             SSIA :)                        */
/**********************************************/
void
calcnewo (struct ShellyArguments *SA, double Oold, double *newO)
{
  double O = Oold, od;
  int mittenodul = NO;
  od = SA->od * pi / 180;

  O += SA->Scano;

  while ((fabs (innodule (O - Oold, Oold)) < SA->heightdiffo) && (O - Oold < od) && (mittenodul == NO))
    {
      O += SA->Scano;
      if ((innodule (SA->Scano, Oold) < 0) && (innodule (SA->Scano, O) > 0))
	{
	  mittenodul = YES;
	}
    }

  *newO = O;

  return;

}				/* calcnewo */


/**********************************************/
/* CalcNewNod:                                */
/*             SSIA :)                        */
/**********************************************/
void
CalcNewNod (struct ShellyArguments *ShellyArgs, char *fout)
{
  struct punkt *list1 = NULL, *list2;

  double O, newO, omin, omax;
  FILE *fptemp, *fp;
  char *hilf, temp[laenge];

  omin = gl.omin;
  omax = gl.omax;
  O = omin;

  mycopystr (fout, temp, '\0');
  hilf = strchr (temp, '\0');
  if (hilf != NULL)
    mycopystr (".tmp", hilf, '\0');


  if ((fptemp = fopen (temp, "w+")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", temp);
      exit (5);
    }

  if ((fp = fopen (fout, "w")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", fout);
      exit (5);
    }
  if ((*ShellyArgs).output == X3D)
    writex3dheader (fp);
  if ((*ShellyArgs).output == T3D)
    writet3dheader (fp);


  fprintf (stdout, "Calculating the Shell:");

  if (calcnodline (omin, ShellyArgs, &list1) != 0)
    exit (5);

  while (O <= omax)
    {
      calcnewo (ShellyArgs, O, &newO);
      O = newO;

      if (calcnodline (O, ShellyArgs, &list2) != 0)
	exit (5);

      fprintf (stdout, ".");
      fflush (stdout);

      if ((*ShellyArgs).output == RPL)
	writerplline (fptemp, list1, list2, 1);
      if ((*ShellyArgs).output == POV)
	writepovtriangles (fptemp, list1, list2, 1);
      if ((*ShellyArgs).output == X3D)
	writex3dline (fptemp, list1, list2, 1);
      if ((*ShellyArgs).output == T3D)
	writet3dline (fptemp, list1, list2, 1);

      if ((*ShellyArgs).output == RAW)
	writerawtriangles (fp, list1, list2, 1);

      myfree (list1);
      list1 = list2;

    }


  if ((*ShellyArgs).output == POV)
    writepovheader (fp);


  if ((*ShellyArgs).output == RPL)
    writerplfinish (fptemp, fp, 1);
  if ((*ShellyArgs).output == POV)
    writepovfinish (fptemp, fp);
  if ((*ShellyArgs).output == X3D)
    writex3dfinish (fptemp, fp, 1);
  if ((*ShellyArgs).output == T3D)
    writet3dfinish (fptemp, fp, 1);

  fprintf (stdout, "\n");

  myfree (list1);
  if ((*ShellyArgs).output == X3D)
    writex3dline (fptemp, NULL, NULL, 2);	/* free temporary pointlist in writex3dline */
  if ((*ShellyArgs).output == T3D)
    writet3dline (fptemp, NULL, NULL, 2);	/* free temporary pointlist in writet3dline */
  if ((*ShellyArgs).output == RPL)
    writerplline (fptemp, NULL, NULL, 2);	/* free temporary pointlist in writet3dline */

  if (fclose (fptemp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", temp);

  if (fclose (fp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", fout);
  remove (temp);
}				/* CalcNewNod */


/**********************************************/
/* CalcNodule:                                */
/*             SSIA :)                        */
/**********************************************/
void
CalcNodule (struct ShellyArguments *ShellyArgs, char *fout)
{
  struct punkt *list1 = NULL, *list2;

  double O, newO, omin, omax;
  FILE *fptemp, *fp;
  char *hilf, temp[laenge];

  omin = gl.omin;
  omax = gl.omax;

  O = omin;

  mycopystr (fout, temp, '\0');
  hilf = strchr (temp, '\0');
  if (hilf != NULL)
    mycopystr (".tmp", hilf, '\0');


  if ((fptemp = fopen (temp, "w+")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", temp);
      exit (5);
    }

  if ((fp = fopen (fout, "w")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", fout);
      exit (5);
    }


  fprintf (stdout, "Calculating the Shell:");

  if (calcline (omin, ShellyArgs, &list1) != 0)
    exit (5);

  if ((*ShellyArgs).output == RPL)
    writerplline (fptemp, list1, NULL, 0);
  if ((*ShellyArgs).output == T3D)
    writet3dline (fptemp, list1, NULL, 0);
  if ((*ShellyArgs).output == X3D)
    writex3dline (fptemp, list1, NULL, 0);


  while (O <= omax)
    {

      calcnewo (ShellyArgs, O, &newO);
      O = newO;

      if (calcline (O, ShellyArgs, &list2) != 0)
	exit (5);

      fprintf (stdout, ".");
      fflush (stdout);

      if ((*ShellyArgs).output == RPL)
	writerplline (fptemp, list1, NULL, 0);
      if ((*ShellyArgs).output == POV)
	writepovtriangles (fptemp, list1, list2, 0);
      if ((*ShellyArgs).output == T3D)
	writet3dline (fptemp, list1, NULL, 0);
      if ((*ShellyArgs).output == X3D)
	writex3dline (fptemp, list1, NULL, 0);
      if ((*ShellyArgs).output == RAW)
	writerawtriangles (fp, list1, list2, 0);

      myfree (list1);
      list1 = list2;
    }

  if ((*ShellyArgs).output == POV)
    writepovheader (fp);
  if ((*ShellyArgs).output == T3D)
    writet3dheader (fp);
  if ((*ShellyArgs).output == X3D)
    writex3dheader (fp);


  if ((*ShellyArgs).output == RPL)
    writerplfinish (fptemp, fp, 0);
  if ((*ShellyArgs).output == T3D)
    writet3dfinish (fptemp, fp, 0);
  if ((*ShellyArgs).output == POV)
    writepovfinish (fptemp, fp);
  if ((*ShellyArgs).output == X3D)
    writex3dfinish (fptemp, fp, 0);

  fprintf (stdout, "\n");

  myfree (list1);


  if (fclose (fptemp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", temp);

  if (fclose (fp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", fout);
  remove (temp);
}				/* CalcNodule */


/**********************************************/
/* CalcShell:                                 */
/*             SSIA :)                        */
/**********************************************/
void
CalcShell (struct ShellyArguments *ShellyArgs, char *fout)
{
  struct punkt *list1 = NULL, *list2;

  double O, omin, omax, od;
  FILE *fptemp, *fp;
  char temp[laenge], *hilf;

  omin = gl.omin;
  omax = gl.omax;
  od = gl.od;


  mycopystr (fout, temp, '\0');
  hilf = strchr (temp, '\0');
  if (hilf != NULL)
    mycopystr (".tmp", hilf, '\0');

  if ((fptemp = fopen (temp, "w+")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", temp);
      exit (5);
    }

  if ((fp = fopen (fout, "w")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", fout);
      exit (5);
    }




  fprintf (stdout, "Calculating the Shell:");

  if (calcline (omin, ShellyArgs, &list1) != 0)
    exit (5);


  if ((*ShellyArgs).output == RPL)
    writerplline (fptemp, list1, NULL, 0);
  if ((*ShellyArgs).output == T3D)
    writet3dline (fptemp, list1, NULL, 0);
  if ((*ShellyArgs).output == X3D)
    writex3dline (fptemp, list1, NULL, 0);

  for (O = omin + od; (O + od) < omax; O = O + od)
    {

      gl.counth = 0;

      if (calcline (O, ShellyArgs, &list2) != 0)
	exit (5);

      fprintf (stdout, ".");
      fflush (stdout);


      if ((*ShellyArgs).output == RPL)
	writerplline (fptemp, list1, NULL, 0);
      if ((*ShellyArgs).output == POV)
	writepovtriangles (fptemp, list1, list2, 0);
      if ((*ShellyArgs).output == T3D)
	writet3dline (fptemp, list1, NULL, 0);
      if ((*ShellyArgs).output == RAW)
	writerawtriangles (fp, list1, list2, 0);
      if ((*ShellyArgs).output == X3D)
	writex3dline (fptemp, list1, NULL, 0);


      myfree (list1);
      list1 = list2;

    }				/* for */

  if ((*ShellyArgs).output == POV)
    writepovheader (fp);
  if ((*ShellyArgs).output == T3D)
    writet3dheader (fp);
  if ((*ShellyArgs).output == X3D)
    writex3dheader (fp);


  if ((*ShellyArgs).output == RPL)
    writerplfinish (fptemp, fp, 0);
  if ((*ShellyArgs).output == T3D)
    writet3dfinish (fptemp, fp, 0);
  if ((*ShellyArgs).output == POV)
    writepovfinish (fptemp, fp);
  if ((*ShellyArgs).output == X3D)
    writex3dfinish (fptemp, fp, 0);

  fprintf (stdout, "\n");

  myfree (list1);


  if (fclose (fptemp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", temp);

  if (fclose (fp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", fout);
  remove (temp);
}				/* CalcShell */


/**********************************************/
/* CalcShell2:                                */
/* It allows use of the BezierPatch capa-     */
/* bilities in POV2.x                         */
/**********************************************/
void
CalcShell2 (struct ShellyArguments *ShellyArgs, char *fout)
{
  struct punkt *list1, *list2, *list3, *list4;

  double O, omin, omax, od;
  FILE *fptemp, *fp;
  char temp[laenge], *hilf;

  omin = gl.omin;
  omax = gl.omax;
  od = gl.od;


  mycopystr (fout, temp, '\0');
  strcat (temp, ".tmp");

  if ((fptemp = fopen (temp, "w+")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", temp);
      exit (5);
    }

  if ((fp = fopen (fout, "w")) == NULL)
    {
      fprintf (stdout, "Could not open outfile: %s!\n", fout);
      exit (5);
    }

  fprintf (stdout, "Calculating the Shell:");


  O = omin;

  if (calcline (O, ShellyArgs, &list1) != 0)
   exit (5);
  O += od;

  while ((O + od) < omax)
    {

      if (calcline (O, ShellyArgs, &list2) != 0)
	exit (5);
      O += od;
      if (calcline (O, ShellyArgs, &list3) != 0)
	exit (5);
      O += od;
      if (calcline (O, ShellyArgs, &list4) != 0)
	exit (5);
      O += od;

      fprintf (stdout, ".");
      fflush (stdout);

      writepovpatches (fptemp, list1, list2, list3, list4);

      myfree (list1);
      myfree (list2);
      myfree (list3);
      list1 = list4;

    }				/* for */

  writepovheader (fp);


  writepovfinish (fptemp, fp);

  fprintf (stdout, "\n");


  if (fclose (fptemp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", temp);

  if (fclose (fp) != 0)
    fprintf (stdout, "Error while closing file: %s!\n", fout);
  remove (temp);
}				/* CalcShell2 */


/**********************************************/
/* RenderPov:                                 */
/*           invoke pov for rendering         */
/**********************************************/
void
RenderPov (struct ShellyArguments *ShellyArgs, char *fout)
{
  char call[laenge];

  printf ("Rendering...\n");

  mycopystr ("pov ", call, '\0');
  mycopystr ((*ShellyArgs).povargs, &call[4], '\0');
  strcat (call, " +i");
  strcat (call, fout);
  system (call);


}				/* RenderPov */


/********************************************/
/*calcnewparams:                            */
/********************************************/
void
calcnewparams (struct ShellyArguments *SAstart, struct ShellyArguments *SAend, struct ShellyArguments *SAi)
{
  static int called = NO;
  static double dalpha = 0, dbeta = 0, dphi = 0, dmy = 0, domega = 0, dA = 0,
    da = 0, db = 0;
  static double domin = 0, domax = 0, dsmin = 0, dsmax = 0, dod = 0, dsd = 0,
    dScale = 0;
  static double dP = 0, dW1 = 0, dW2 = 0, dN = 0, dL = 0;
  static double dP2 = 0, dW12 = 0, dW22 = 0, dN2 = 0, dL2 = 0, dOff2 = 0;
  static double dP3 = 0, dW13 = 0, dW23 = 0, dN3 = 0, dL3 = 0, dOff3 = 0;
  static double dcamx = 0, dcamy = 0, dcamz = 0;


  if (called == NO)
    {
      if (SAstart->alpha != SAend->alpha)
	dalpha = ((SAend->alpha - SAstart->alpha) / SAstart->steps);
      if (SAstart->beta != SAend->beta)
	dbeta = ((SAend->beta - SAstart->beta) / SAstart->steps);
      if (SAstart->phi != SAend->phi)
	dphi = ((SAend->phi - SAstart->phi) / SAstart->steps);
      if (SAstart->my != SAend->my)
	dmy = ((SAend->my - SAstart->my) / SAstart->steps);
      if (SAstart->omega != SAend->omega)
	domega = ((SAend->omega - SAstart->omega) / SAstart->steps);
      if (SAstart->A != SAend->A)
	dA = ((SAend->A - SAstart->A) / SAstart->steps);
      if (SAstart->a != SAend->a)
	da = ((SAend->a - SAstart->a) / SAstart->steps);
      if (SAstart->b != SAend->b)
	db = ((SAend->b - SAstart->b) / SAstart->steps);
      if (SAstart->omin != SAend->omin)
	domin = ((SAend->omin - SAstart->omin) / SAstart->steps);
      if (SAstart->omax != SAend->omax)
	domax = ((SAend->omax - SAstart->omax) / SAstart->steps);
      if (SAstart->smin != SAend->smin)
	dsmin = ((SAend->smin - SAstart->smin) / SAstart->steps);
      if (SAstart->smax != SAend->smax)
	dsmax = ((SAend->smax - SAstart->smax) / SAstart->steps);
      if (SAstart->od != SAend->od)
	dod = ((SAend->od - SAstart->od) / SAstart->steps);
      if (SAstart->sd != SAend->sd)
	dsd = ((SAend->sd - SAstart->sd) / SAstart->steps);
      if (SAstart->Scale != SAend->Scale)
	dScale = ((SAend->Scale - SAstart->Scale) / SAstart->steps);
      if (SAstart->P != SAend->P)
	dP = ((SAend->P - SAstart->P) / SAstart->steps);
      if (SAstart->L != SAend->L)
	dL = ((SAend->L - SAstart->L) / SAstart->steps);
      if (SAstart->N != SAend->N)
	dN = ((SAend->N - SAstart->N) / SAstart->steps);
      if (SAstart->W1 != SAend->W1)
	dW1 = ((SAend->W1 - SAstart->W1) / SAstart->steps);
      if (SAstart->W2 != SAend->W2)
	dW2 = ((SAend->W2 - SAstart->W2) / SAstart->steps);
      if (SAstart->P2 != SAend->P2)
	dP2 = ((SAend->P2 - SAstart->P2) / SAstart->steps);
      if (SAstart->L2 != SAend->L2)
	dL2 = ((SAend->L2 - SAstart->L2) / SAstart->steps);
      if (SAstart->N2 != SAend->N2)
	dN2 = ((SAend->N2 - SAstart->N2) / SAstart->steps);
      if (SAstart->Off2 != SAend->Off2)
	dOff2 = ((SAend->Off2 - SAstart->Off2) / SAstart->steps);
      if (SAstart->W12 != SAend->W12)
	dW12 = ((SAend->W12 - SAstart->W12) / SAstart->steps);
      if (SAstart->W22 != SAend->W22)
	dW22 = ((SAend->W22 - SAstart->W22) / SAstart->steps);
      if (SAstart->P3 != SAend->P3)
	dP3 = ((SAend->P3 - SAstart->P3) / SAstart->steps);
      if (SAstart->L3 != SAend->L3)
	dL3 = ((SAend->L3 - SAstart->L3) / SAstart->steps);
      if (SAstart->N3 != SAend->N3)
	dN3 = ((SAend->N3 - SAstart->N3) / SAstart->steps);
      if (SAstart->Off3 != SAend->Off3)
	dOff3 = ((SAend->Off3 - SAstart->Off3) / SAstart->steps);
      if (SAstart->W13 != SAend->W13)
	dW13 = ((SAend->W13 - SAstart->W13) / SAstart->steps);
      if (SAstart->W23 != SAend->W23)
	dW23 = ((SAend->W23 - SAstart->W23) / SAstart->steps);
      if (SAstart->camx != SAend->camx)
	dcamx = ((SAend->camx - SAstart->camx) / SAstart->steps);
      if (SAstart->camy != SAend->camy)
	dcamy = ((SAend->camy - SAstart->camy) / SAstart->steps);
      if (SAstart->camz != SAend->camz)
	dcamz = ((SAend->camz - SAstart->camz) / SAstart->steps);

      called = YES;
    }

  SAi->alpha = SAi->alpha + dalpha;
  SAi->phi = SAi->phi + dphi;
  SAi->my = SAi->my + dmy;
  SAi->omega = SAi->omega + domega;
  SAi->omin = SAi->omin + domin;
  SAi->omax = SAi->omax + domax;
  SAi->smin = SAi->smin + dsmin;
  SAi->smax = SAi->smax + dsmax;
  SAi->od = SAi->od + dod;
  SAi->sd = SAi->sd + dsd;
  SAi->Scale = SAi->Scale + dScale;
  SAi->A = SAi->A + dA;
  SAi->a = SAi->a + da;
  SAi->b = SAi->b + db;
  SAi->P = SAi->P + dP;
  SAi->W1 = SAi->W1 + dW1;
  SAi->W2 = SAi->W2 + dW2;
  SAi->N = SAi->N + dN;
  SAi->L = SAi->L + dL;
  SAi->P2 = SAi->P2 + dP2;
  SAi->W12 = SAi->W12 + dW12;
  SAi->W22 = SAi->W22 + dW22;
  SAi->N2 = SAi->N2 + dN2;
  SAi->L2 = SAi->L2 + dL2;
  SAi->Off2 = SAi->Off2 + dOff2;
  SAi->P3 = SAi->P3 + dP3;
  SAi->W13 = SAi->W13 + dW13;
  SAi->W23 = SAi->W23 + dW23;
  SAi->N3 = SAi->N3 + dN3;
  SAi->L3 = SAi->L3 + dL3;
  SAi->Off3 = SAi->Off3 + dOff3;
  SAi->camx = SAi->camx + dcamx;
  SAi->camy = SAi->camy + dcamy;
  SAi->camz = SAi->camz + dcamz;


}				/* calcnewparams */


/********************************************/
/*animate:                                  */
/********************************************/
void
animate (struct ShellyArguments *SAstart, int ac, char **av)
{
  int i;
  char fn[laenge];
  struct ShellyArguments *SAend, SAi;



  if (SAstart == NULL)
    return;

  mycopystr (av[2], fn, '\0');

  if ((SAend = calloc (1, sizeof (struct ShellyArguments))) != NULL);
  {
    ReadInfile (SAend, SAstart->animto);
    printf ("Doing Animation, %d steps.\n", SAstart->steps);

    SAstart->steps--;
    SAi = *SAstart;

    for (i = 0; i <= SAstart->steps; i++)
      {
	mycopystr (av[2], fn, '\0');
	sprintf (fn, "%s%d", fn, i);

	init (&SAi);

        if (SAstart->output == BEZ)
         CalcShell2(&SAi, fn); 

	if (SAi.mode == NORMAL)
	  CalcShell (&SAi, fn);
	if (SAi.mode == NODULE)
	  CalcNodule (&SAi, fn);
	if (SAi.mode == NEWNOD)
	  CalcNewNod (&SAi, fn);

	if (SAi.Render == YES)
	  if (SAi.output != POV)
	    fprintf (stdout, "Rendering is only available for POV-output!\n(Ever tried to feed a RPL-Macro into POV? ;))\n");
	  else
	    RenderPov (&SAi, fn);

	calcnewparams (SAstart, SAend, &SAi);

      }				/* for */

  }
  exit (0);
}

/********************************************/
/*main:                                     */
/********************************************/
int
main (int ac, char **av)
{
  struct ShellyArguments *ShellyArgs;

  fprintf (stdout, "ShellyV1.6, the ShellShapeGenerator by RANDi\n");

  if (ac != 3)
    {
      fprintf (stdout, "USAGE:\n\n");
      fprintf (stdout, "   'shelly infilename outfilename'\n\n");
      exit (1);
    }				/* if */


  if ((ShellyArgs = calloc (1, sizeof (struct ShellyArguments))) != NULL);
  {
    ReadInfile (ShellyArgs, av[1]);

    init (ShellyArgs);

    if (ShellyArgs->anim == YES)
      animate (ShellyArgs, ac, av);

    if (ShellyArgs->output == BEZ)
      CalcShell2(ShellyArgs, av[2]); 
    else
    {
     if (ShellyArgs->mode == NORMAL)
       CalcShell (ShellyArgs, av[2]);
     if (ShellyArgs->mode == NODULE)
       CalcNodule (ShellyArgs, av[2]);
     if (ShellyArgs->mode == NEWNOD)
       CalcNewNod (ShellyArgs, av[2]);
    }
    
    if (ShellyArgs->Render == YES)
      if (ShellyArgs->output != POV || ShellyArgs->output != BEZ)
	fprintf (stdout, "Rendering is only available for POV-output!\n(Ever tried to feed a RPL-Macro into POV? ;))\n");
      else
	RenderPov (ShellyArgs, av[2]);

    free (ShellyArgs);
  }
  exit (0);
}
